package edu.caltech.sbml;

/*
** Filename    : SBMLValidate.java
** Description : application class for the SBML validator/NOM module + top level of NOMService implementation
** Author(s)   : SBW Development Group <sysbio-team@caltech.edu>
** Organization: Caltech ERATO Kitano Systems Biology Project
** Created     : 2001-07-07
** Revision    : $Id: TJXML.java,v 1.2 2004/06/07 23:51:15 sramsey Exp $
** $Source: /proj/bolrilab/CVS_REPOSITORY/bolouri/java/extsrc/SBMLReader/TJXML.java,v $
**
** Copyright 2001 California Institute of Technology and
** Japan Science and Technology Corporation.
**
** This library is free software; you can redistribute it and/or modify it
** under the terms of the GNU Lesser General Public License as published
** by the Free Software Foundation; either version 2.1 of the License, or
** any later version.
**
** This library is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
** MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
** documentation provided hereunder is on an "as is" basis, and the
** California Institute of Technology and Japan Science and Technology
** Corporation have no obligations to provide maintenance, support,
** updates, enhancements or modifications.  In no event shall the
** California Institute of Technology or the Japan Science and Technology
** Corporation be liable to any party for direct, indirect, special,
** incidental or consequential damages, including lost profits, arising
** out of the use of this software and its documentation, even if the
** California Institute of Technology and/or Japan Science and Technology
** Corporation have been advised of the possibility of such damage.  See
** the GNU Lesser General Public License for more details.
**
** You should have received a copy of the GNU Lesser General Public License
** along with this library; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
**
** The original code contained here was initially developed by:
**
**     Andrew Finney, Herbert Sauro, Michael Hucka, Hamid Bolouri
**     The Systems Biology Workbench Development Group
**     ERATO Kitano Systems Biology Project
**     Control and Dynamical Systems, MC 107-81
**     California Institute of Technology
**     Pasadena, CA, 91125, USA
**
**     http://www.cds.caltech.edu/erato
**     mailto:sysbio-team@caltech.edu
**
** Contributor(s):
**
** Contributor(s):
**
** sramsey  2003/08/28  Remove dependence on Xerces library.  Using Crimson
**                      library instead, because it is the default XML parser
**                      that ships with the Sun JDK 1.4.2
**
** sramsey  2003/10/20  Modified to read SBML Level 1, Version 2 documents,
**                      including "Rules", which were introduced in SBML L1V2.
**
** sramsey  2004/02/13  Changed to "edu.caltech.sbml" package; unified
**                      all the source code in this directory into a single
**                      package namespace.
*/

import java.io.*;
import java.lang.String.*;
import javax.swing.*;
import java.net.*;

import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.xml.sax.*;

import edu.caltech.sbw.*;

// TJXML is a XML decorator class for the NOM object class
// The NOM object class defines the interface to an internal representation
// of cellular models. The XML decorator class adds SBML import/export
// functionality to the NOM object

class TJXML implements ErrorHandler {

    private static final String NODE_TAG_SPECIES_RULE = "speciesConcentrationRule";
    private static final String NODE_TAG_PARAMETER_RULE = "parameterRule";
    private static final String NODE_TAG_COMPARTMENT_RULE = "compartmentVolumeRule";
    private static final String ATTRIBUTE_SPECIES = "species";
    private static final String ATTRIBUTE_PARAMETER = "name";
    private static final String ATTRIBUTE_COMPARTMENT = "compartment";
    private static final String ATTRIBUTE_FORMULA = "formula";
    private static final String ATTRIBUTE_TYPE = "type";

    // Depending on which constructor is called, the network variable
    // can either be owned by a TJXML instance or passed to it from
    // an outside source.

    // Network object, if the Network object is managed by an external source
    // then this variable is a reference to an external instance
    TJNetwork Network;

    // This is the constructor which results in TJXML managing its own Network object
    TJXML() {
      Network = new TJNetwork();
    }

    /**
     * error string generated by parser
     */
    private String error ;

    // ------------------------------------------------------------------------
    // Get an attribute with name 'AttrName' from a given node. The argument
    // DefaultStr provide a default result if 'AttrName' cannot be found.
    // ------------------------------------------------------------------------

    String GetAttribute (Node node, String AttrName, String DefaultStr) {
         String result = DefaultStr;
         NamedNodeMap Attr = node.getAttributes();
         if (Attr != null) {
            Node AttrNode = Attr.getNamedItem (AttrName);
            if (AttrNode != null)
               result = AttrNode.getNodeValue();
         }
         return result;
    }

    // ------------------------------------------------------------------------
    // Returns a DOMParser. The reason for doing this is because the means by
    // which Parsers are created is not specifed in the W3C docs and each DOM
    // implementation has it's own approach. If we switch to a different DOM
    // library, in theory we need only change this bit.
    // ------------------------------------------------------------------------
    DocumentBuilder GetXMLParser() throws FactoryConfigurationError, ParserConfigurationException {
        DocumentBuilder result = DocumentBuilderFactory.newInstance().newDocumentBuilder();
//         File classFile = Sys.getFileLoadedForClass(SBMLValidate.class);

//         result.setFeature( "http://xml.org/sax/features/validation",
//                            true );
//         result.setFeature( "http://xml.org/sax/features/namespaces",
//                            true );
//         result.setFeature( "http://apache.org/xml/features/validation/schema",
//                            true );
//         result.setFeature( "http://apache.org/xml/features/validation/schema-full-checking",
//                            true );

//         // Fix for bug #713619.
//         // Ugly hack.  Because we mustn't have spaces in the URL for the
//         // schema file here, and because Windows users want to install SBW
//         // into "\Program Files\something" (which does have spaces), I see
//         // no alternative but to put the schema file into someplace
//         // guaranteed not to have spaces in the name.  That only seems to
//         // be the Windows system directory.  The nasty part here is that
//         // we can't really be sure that the pathname doesn't have spaces
//         // on non-Windows systems.  This needs to be revisited in the future.

//         String pathToSchema;
//         if (Sys.OSIsWindows())
//             pathToSchema = Sys.getWindowsSystemDirectory()
//                 + File.separator + "sbml.xsd";
//         else
//             pathToSchema = classFile.getParentFile().toURL().toString()
//                 + "sbml.xsd";

//         result.setProperty(
//             "http://apache.org/xml/properties/schema/external-schemaLocation",
//             "http://www.sbml.org/sbml/level1 " + pathToSchema);

         error = "";
         result.setErrorHandler(this);

        return result;
    }

    // -------------------------
    // XML-Scheme error handlers
    // -------------------------

    // Warning
    public void warning(SAXParseException ex) {
        System.err.println("[Warning] "+
                           getLocationString(ex)+": "+
                           ex.getMessage());
    }

    // Error
    public void error(SAXParseException ex)
    {
        String errorMsg = "[Error] "+ getLocationString(ex)+": "+ ex.getMessage() + '\n';

        error += errorMsg ;
    }

    // Fatal error
    public void fatalError(SAXParseException ex)
    {
        String errorMsg = "[Error] "+ getLocationString(ex)+": "+ ex.getMessage() + '\n';

        error += errorMsg ;
    }

    // Returns a string of the location
    private String getLocationString(SAXParseException ex) {
        StringBuffer str = new StringBuffer();

        String systemId = ex.getSystemId();
        if (systemId != null) {
            int index = systemId.lastIndexOf('/');
            if (index != -1)
                systemId = systemId.substring(index + 1);
            str.append(systemId);
        }
        str.append(':');
        str.append(ex.getLineNumber());
        str.append(':');
        str.append(ex.getColumnNumber());

        return str.toString();
    }


    // ------------------------------------------------------------------------
    // Returns Fixed or Float according to the species type in Node
    // ------------------------------------------------------------------------

    int GetSpeciesType (Node node) {
       int result = TConstants.jmsFloat; // Default
       NamedNodeMap Attr = node.getAttributes();
       if (Attr != null) {
          Node AttrNode = Attr.getNamedItem("boundaryCondition");
          if (AttrNode != null) {
             String xx = AttrNode.getNodeValue();
             if (AttrNode.getNodeValue().toUpperCase().equals ("FALSE"))
                result = TConstants.jmsFloat;
             else result = TConstants.jmsFixed;
          }
       }
       return result;
    }

    void GetRuleList(TRuleList pRuleList, NodeList pNodeList) throws Exception
    {
        Node node;
        for(int i = 0; i < pNodeList.getLength(); ++i)
        {
            node = pNodeList.item(i);
            if(node.getNodeType() == Node.ELEMENT_NODE)
            {
                String ruleName = null;
                String nodeTag = node.getNodeName();
                TRule.Type ruleType = null;
                String ruleScalar = GetAttribute(node, ATTRIBUTE_TYPE, null);
                if(null != ruleScalar && ruleScalar.equals("rate"))
                {
                    throw new Exception("cannot process a model with a rule whose type is \"rate\"; rule name is: " + ruleName);
                }

                String ruleFormula = GetAttribute(node, ATTRIBUTE_FORMULA, null);
                if(null == ruleFormula)
                {
                    throw new Exception("required attribute not found: " + ATTRIBUTE_FORMULA);
                }

                if(nodeTag.equals(NODE_TAG_SPECIES_RULE))
                {
                    ruleName = GetAttribute(node, ATTRIBUTE_SPECIES, null);
                    if(null == ruleName)
                    {
                        throw new Exception("required attribute not found: " + ATTRIBUTE_SPECIES);
                    }
                    ruleType = TRule.Type.SPECIES;
                }
                else if(nodeTag.equals(NODE_TAG_COMPARTMENT_RULE))
                {
                    ruleName = GetAttribute(node, ATTRIBUTE_COMPARTMENT, null);
                    if(null == ruleName)
                    {
                        throw new Exception("required attribute not found: " + ATTRIBUTE_COMPARTMENT);
                    }
                    ruleType = TRule.Type.COMPARTMENT;
                }
                else if(nodeTag.equals(NODE_TAG_PARAMETER_RULE))
                {
                    ruleName = GetAttribute(node, ATTRIBUTE_PARAMETER, null);
                    if(null == ruleName)
                    {
                        throw new Exception("required attribute not found: " + ATTRIBUTE_PARAMETER);
                    }
                    ruleType = TRule.Type.PARAMETER;
                }
                else
                {
                    throw new Exception ("Syntax error: unknown tag inside listOfRules element: " + nodeTag);
                }
                pRuleList.add(ruleName, ruleFormula, ruleType);
            }
        }
    }

    // ------------------------------------------------------------------------
    // Reads in the global parameter list. The names of the parameters and their
    // values are stored in the ParameterList argument.
    // ------------------------------------------------------------------------

    void GetGlobalParameterList (TParameterList ParameterList, NodeList nodelist) {
         Node node; String Name; String ValueStr; double Value = 0;
         for (int i = 0; i<nodelist.getLength(); i++) {
             node = nodelist.item(i);
             if (node.getNodeType() != node.TEXT_NODE) {
                Name = GetAttribute (node, "name", "");

                int defined ;
                ValueStr = GetAttribute (nodelist.item(i), "value", "");

                if (ValueStr.length() == 0)
                    defined = TConstants.nsUnDefined ;
                else
                {
                    Value = TUtils.StrToFloat(ValueStr);
                    defined = TConstants.nsDefined;
                }

                ParameterList.add (new TParameter(Name, Value, defined));
             }
         }
    }

    // ------------------------------------------------------------------------
    // Retrieves the stoichiometry from node. This is not complete, it doesn't
    // deal with the possible denomiator part of a stoichiometric spec. Should
    // be easy to put in though
    // ------------------------------------------------------------------------

    int GetStoichiometry (Node node) {
        int result = 1;
        NamedNodeMap Attr = node.getAttributes();
        if (Attr != null) {
           Node AttrNode = Attr.getNamedItem("stoichiometry");
           if (AttrNode != null) {
              String S = AttrNode.getNodeValue();
              int p = S.indexOf ("/");
              if (p == -1)
                 result = TUtils.StrToInt (S);
              // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
              // Needs to account for denominator stoichiometry value !!!!!
              //else result = TUtils.StrToInt (
           }
        }
      return result;
    }

    // ------------------------------------------------------------------------
    // Retrieves the list of reactants and stores them in JReaction
    // ------------------------------------------------------------------------

    void GetReactants (TJNetwork Network, TReaction JReaction, NodeList nodelist) {
         Node node; TSpecies Species;
         for (int i = 0; i<nodelist.getLength(); i++) {
             node = nodelist.item(i);
             if (node.getNodeType() != node.TEXT_NODE) {
                String Name = GetAttribute (node, "specie", "");
                if(0 == Name.length())
                {
                    Name = GetAttribute (node, "species", "");
                }

                int stoich = GetStoichiometry (node);
                Species = Network.FindMetabolite (Name);
                if (Species != null) {
                   TReactant JReactant = new TReactant();
                   JReactant.Species = Species;
                   JReactant.Stoichiometry = stoich;
                   JReaction.ReactantList.add (JReactant);
                }
             }

         }
    }

    // ------------------------------------------------------------------------
    // Retrieves the list of products and stores them in JReaction
    // ------------------------------------------------------------------------

    void GetProducts (TJNetwork Network, TReaction JReaction, NodeList nodelist) {
         Node node; TSpecies Species;
         for (int i = 0; i<nodelist.getLength(); i++) {
             node = nodelist.item(i);
             if (node.getNodeType() != node.TEXT_NODE) {
                String Name = GetAttribute (node, "specie", "");
                if(0 == Name.length())
                {
                    Name = GetAttribute(node, "species", "");
                }
                int stoich = GetStoichiometry (node);
                Species = Network.FindMetabolite (Name);
                if (Species != null) {
                   TReactant JReactant = new TReactant();
                   JReactant.Species = Species;
                   JReactant.Stoichiometry = stoich;
                   JReaction.ProductList.add (JReactant);
                }
             }
         }
    }

    // ------------------------------------------------------------------------
    // Gets the rate law formula attribute and stores it in JReaction
    // Also looks for any local parameters and stores them away to.
    // ------------------------------------------------------------------------

    void GetRateLaw (TReaction JReaction, Node node) {
         JReaction.RateLaw.expression = GetAttribute (node, "formula", "");
         NodeList ElementList = node.getChildNodes();
         int length = ElementList.getLength();

         String name;
         for (int i=0; i<length; i++) {
             Node inode = ElementList.item(i);
             if (inode.getNodeType() != inode.TEXT_NODE) {
                 name = inode.getNodeName();
                 if (name.equals("listOfParameters")) {
                    GetParameters (JReaction, inode.getChildNodes());
                    break;
                 }
             }
         }
    }

    String getSubstanceUnitsString(NodeList nodeList) throws Exception
    {
        Node node;
        int numNodes = nodeList.getLength();
        StringBuffer unitsStringBuf = new StringBuffer();
        String scaleString;
        boolean firstUnit = true;
        String name;
        for(int i = 0; i < numNodes; ++i)
        {
            node = nodeList.item(i);
            if(node.getNodeType() == Node.TEXT_NODE)
            {
                continue;
            }
            assert (node.getNodeName().equals("unitDefinition")) : "unexpected element instead of unitDefinition: " + node.getNodeName();
            name = GetAttribute(node, "name", "");
            if(name.length() != 0)
            {
                if(name.equals("substance"))
                {
                    NodeList unitDefinitionSubstanceChildren = node.getChildNodes();
                    int numUnitDefinitionSubstanceChildren = unitDefinitionSubstanceChildren.getLength();
                    for(int k = 0; k < numUnitDefinitionSubstanceChildren; ++k)
                    {
                        Node unitDefinitionSubstanceChild = unitDefinitionSubstanceChildren.item(k);
                        if(unitDefinitionSubstanceChild.getNodeType() != Node.TEXT_NODE)
                        {
                            assert (unitDefinitionSubstanceChild.getNodeName().equals("listOfUnits")) : "unexpected element instead of listOfUnits: " + unitDefinitionSubstanceChild.getNodeName();
                            NodeList childNodes = unitDefinitionSubstanceChild.getChildNodes();
                            int numChildNodes = childNodes.getLength();
                            Node childNode = null;
                            for(int j = 0; j < numChildNodes; ++j)
                            {
                                childNode = childNodes.item(j);
                                if(childNode.getNodeType() != Node.TEXT_NODE)
                                {
                                    if(null != childNode)
                                    {
                                        name = GetAttribute(childNode, "kind", "");
                                        if(name.length() > 0)
                                        {
                                            int scale = 1;
                                            scaleString = GetAttribute(childNode, "scale", "");
                                            if(scaleString.length() > 0)
                                            {
                                                scale = Integer.parseInt(scaleString);
                                            }
                                            if(firstUnit)
                                            {
                                                firstUnit = false;
                                            }
                                            else
                                            {
                                                unitsStringBuf.append(" * ");
                                            }
                                            unitsStringBuf.append(name.trim());
                                            if(scale != 1)
                                            {
                                                unitsStringBuf.append("^(" + scale + ")");
                                            }
                                        }
                                        else
                                        {
                                            throw new Exception("required attribute \"kind\" not supplied, for \"unit\" element");
                                        }
                                    }    
                                }    
                            }                   
                            
                        }
                    }
                }
            }
        }

        String unitsString = null;

        if(unitsStringBuf.length() > 0)
        {
            unitsString = unitsStringBuf.toString();
        }

        return(unitsString);
    }


    // ------------------------------------------------------------------------
    // Stores any local parameters associated with the current rate law
    // ------------------------------------------------------------------------

    void GetParameters (TReaction JReaction, NodeList nodelist) {
          Node node; String name; double value = 0; TParameterList plist;
          for (int i = 0; i<nodelist.getLength(); i++) {
               node = nodelist.item(i);
               if (node.getNodeType() != node.TEXT_NODE) {
                  name = GetAttribute (node, "name", "");

                  int defined ;
                  String valueStr = GetAttribute (node, "value", "");

                  if (valueStr.length() == 0)
                      defined = TConstants.nsUnDefined;
                  else
                  {
                      defined = TConstants.nsDefined;
                      value = TUtils.StrToFloat(valueStr);
                  }

                  plist = JReaction.RateLaw.ParameterList;
                  plist.add (new TParameter(name, value, defined));
               }
          }
    }



    // ------------------------------------------------------------------------
    // Given a Document (a pre-loaded XML script), inspect the DOM tree and
    // copy the model into the NOM object
    // ------------------------------------------------------------------------

    void BuildNetworkModel (Document doc) throws Exception {
        Node node;

        // Read in the comnpartment information

        Network.Name = GetAttribute (doc.getElementsByTagName("model").item(0), "name", "untitled");
        if (doc.getElementsByTagName("listOfCompartments").item(0) != null) {
           NodeList nodeList = doc.getElementsByTagName("listOfCompartments").item(0).getChildNodes();
           for (int i=0; i<nodeList.getLength(); i++) {
               node = nodeList.item (i);
               if (node.getNodeType() != node.TEXT_NODE) {
                  if (!node.getNodeName().equals("compartment"))
                     throw new Exception ("Syntax error: expecting 'compartment' tag in list of compartments");

                  String Name = GetAttribute (node, "name", "");
                  String VolumeStr = GetAttribute (node, "volume", "1.0");
                  if (Name.length() != 0)
                      Network.VolumeList.add (Name, TUtils.StrToFloat (VolumeStr));
              }
           }
        }

        // Read in the species information

        if (doc.getElementsByTagName ("listOfSpecies") != null) {
            NodeList nodelist = doc.getElementsByTagName ("listOfSpecies").item(0).getChildNodes();
            for (int i=0; i<nodelist.getLength(); i++) {
                node = nodelist.item(i);
                if (node.getNodeType() != node.TEXT_NODE) {
                   if (!node.getNodeName().equals("specie") &&
                       !node.getNodeName().equals("species"))
                      throw new Exception ("Syntax error: expecting 'specie' tag in list of species");

                   int defined ;
                   String valueStr = GetAttribute (node, "initialAmount", "");
                   double value = 0;

                   if (valueStr.length() == 0)
                       defined = TConstants.nsUnDefined;
                   else
                   {
                        value = TUtils.StrToFloat(valueStr);
                        defined = TConstants.nsDefined ;
                   }

                   String VolumeStr = GetAttribute (node, "compartment", "");
                   String name = GetAttribute (node, "name", "");

                   if (GetSpeciesType (node) == TConstants.jmsFloat) {
                      int SIndex = Network.MetaboliteList.add (name, TConstants.jmsFloat, defined, value);

                      // If the species has a volume element then find the volume object
                      // and associate it with the species

                      // Create an integer object so that we can return integer
                      // values from a method argument list
                      IntObj VolIndex = new IntObj();

                      if (VolumeStr != "") {
                         if (Network.VolumeList.find (VolumeStr, VolIndex)) {
                            Network.MetaboliteList.get(SIndex).Volume = Network.VolumeList.get (VolIndex.i);
                         } else {
                            throw new SBWApplicationException(VolumeStr + "is not a declared volume on species " + name, "");
                         }
                      }
                   } else {
                      int SIndex = Network.BoundaryList.add (name, TConstants.jmsFixed, defined, value);

                      IntObj VolIndex = new IntObj();

                      if (VolumeStr != "") {
                         if (Network.VolumeList.find (VolumeStr, VolIndex)) {
                            Network.BoundaryList.get(SIndex).Volume = Network.VolumeList.get (VolIndex.i);
                         } else {
                            throw new SBWApplicationException(VolumeStr + "is not a declared volume on species " + name, "");
                         }
                      }
                  }
              }
            }  // End of listOfSpecies

          // Read in the global parameters

          node = doc.getElementsByTagName ("listOfParameters").item(0);
          if (node != null && node.getParentNode().getNodeName().equals("model"))
             GetGlobalParameterList (Network.GlobalParameterList, node.getChildNodes());

          node = doc.getElementsByTagName("listOfRules").item(0);
          if (node != null)
          {
              TRuleList ruleList = Network.RuleList;
              GetRuleList(ruleList, node.getChildNodes());
              // process list of rules
          }

          node = doc.getElementsByTagName("listOfUnitDefinitions").item(0);
          if(null != node)
          {
              assert (node.getNodeName().equals("listOfUnitDefinitions")) : "unexpected element instead of listOfUnitDefinitions: " + node.getNodeName();
              String substanceUnitsString = getSubstanceUnitsString(node.getChildNodes());
              Network.SubstanceUnitsString = substanceUnitsString;
          }
          
          // Read in the reaction specifications

          if (doc.getElementsByTagName("listOfReactions") != null) {
              // Get the list of reactions
              nodelist = doc.getElementsByTagName("listOfReactions").item(0).getChildNodes();
              int length = nodelist.getLength();  // A bit of optimization
              for (int i = 0; i<length; i++) {
                  // Get a single reaction
                  node = nodelist.item(i);

                  // Ignore any nodes that are text nodes
                  if (node.getNodeType() != node.TEXT_NODE) {

                      if (!node.getNodeName().equals("reaction"))
                         throw new Exception ("Syntax error: expecting 'reaction' tag in list of reactions");

                     // First get hold of the reaction name and reversibility flag
                     TReaction JReaction = new TReaction(GetAttribute (node, "name", "_J" + TUtils.IntToStr (i)));
                     if (GetAttribute (node, "reversible", "true").equals("true"))
                        JReaction.Reversible = true;
                     else JReaction.Reversible = false;

                     // Get the list containing the reactants, products and kinetic law
                     NodeList ElementList = node.getChildNodes();

                     for (int j = 0; j<ElementList.getLength(); j++) {
                         node = ElementList.item(j);
                         // Ignore text nodes
                         if (node.getNodeType() != node.TEXT_NODE) {
                            String name = node.getNodeName();
                            if (name.equals("listOfReactants"))
                               GetReactants (Network, JReaction, ElementList.item(j).getChildNodes());
                            if (name.equals("listOfProducts"))
                               GetProducts (Network, JReaction, ElementList.item(j).getChildNodes());
                            if (name.equals("kineticLaw")) {
                               GetRateLaw (JReaction, ElementList.item(j));
                            }
                         }
                     } // end for loop
                     // Finally add the reaction object to the reaction list
                     Network.ReactionList.add (JReaction);
                  }
              }
          }
        }
    }

    // ------------------------------------------------------------------------
    // Load some XML from a given file
    // ------------------------------------------------------------------------

    public void LoadFromFile(String XMLFile) throws SBWException
    {
        try
        {
            DocumentBuilder parser = GetXMLParser();
            Document document = parser.parse(XMLFile);
            postParseProcessing(document);
        }
        catch (Throwable t)
        {
            throw SBWException.translateException(t);
        }
    }

    private void postParseProcessing(Document pDocument) throws Throwable
    {
        if (error.length() != 0)
            throw new SBWApplicationException("Error parsing SBML", error);

        BuildNetworkModel (pDocument);
    }

    // ------------------------------------------------------------------------
    // Load some XML from a given string
    // ------------------------------------------------------------------------

    public void LoadFromString (String XMLStr) throws SBWException
    {
        try {
            DocumentBuilder parser = GetXMLParser();
            InputSource is = new InputSource();

            is.setCharacterStream (new StringReader (XMLStr));
            Document document = parser.parse(is);

            postParseProcessing(document);
        }
        catch (Throwable e) {
            throw SBWException.translateException(e);
        }
    }
}


